from __future__ import annotations
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ylz_utils.langchain.graph.self_rag_graph import SelfRagGraph

from langchain_core.prompts import ChatPromptTemplate
from langchain_core.output_parsers import StrOutputParser

class TransformQueryNode():
    def __init__(self,self_rag_graph:SelfRagGraph):
        self.self_rag_graph = self_rag_graph

    def transform_query(self,state):
        """
        Transform the query to produce a better question.

        Args:
            state (dict): The current graph state

        Returns:
            state (dict): Updates question key with a re-phrased question
        """

        print("---TRANSFORM QUERY---")
        question = state["question"]
        documents = state["documents"]

        # Prompt
        system = """You a question re-writer that converts an input question to a better version that is optimized \n 
            for vectorstore retrieval. Look at the input and try to reason about the underlying semantic intent / meaning."""
        re_write_prompt = ChatPromptTemplate.from_messages(
            [
                ("system", system),
                (
                    "human",
                    "Here is the initial question: \n\n {question} \n Formulate an improved question.",
                ),
            ]
        )

        question_rewriter = re_write_prompt | self.self_rag_graph.llm | StrOutputParser()
        question_rewriter.invoke({"question": question})

        # Re-write question
        better_question = question_rewriter.invoke({"question": question})
        return {"documents": documents, "question": better_question}




from __future__ import annotations
from typing import TYPE_CHECKING, Literal, Optional

if TYPE_CHECKING:
    from ylz_utils.langchain import LangchainLib

from tqdm import tqdm

from ylz_utils.langchain.vectorstores.elasticsearch import ESLib
from ylz_utils.langchain.vectorstores.faiss import FaissLib
from ylz_utils.langchain.vectorstores.chroma import ChromaLib


class VectorstoreLib():
    def __init__(self,langchainLib:LangchainLib,provider:Optional[Literal['es','faiss','chroma']]=None):
        self.langchainLib = langchainLib
        self.config = langchainLib.config
        self.faissLib = FaissLib(self)
        self.esLib = ESLib(self)
        self.chromaLib = ChromaLib(self)
        if provider:
            self.provider = provider
        else:
            self.provider = self.config.get("VECTORSTORE.DEFAULT")

    def get_store(self,provider:Optional[Literal['es','faiss','chroma']],collection_name=None,embedding=None):
        if not provider:
            provider = self.provider
        if not provider:
            raise Exception("请设置vectorstore provider")
        
        if provider=='es':
            return self.esLib.get_store(collection_name,embedding)
        elif provider == 'faiss':
            return self.faissLib.get_store(collection_name,embedding)
        elif provider == 'chroma':
            return self.chromaLib.get_store(collection_name,embedding)
        else:
            return None
    def add_docs(self,vector_store,docs,batch=1) -> list[str]:
        return self._split_batch_and_add(docs,batch,vector_store.add_documents)
    def add_texts(self,vector_store,texts,batch=1) -> list[str]:
        return self._split_batch_and_add(texts,batch,vector_store.add_texts)
    
    def search(self,query,vectorstore,k=10,filter={}):
        return vectorstore.similarity_search(query,k=k,filter=filter)
    
    def search_with_score(self,query,vectorstore,k=10,filter={}):
        return vectorstore.similarity_search_with_score(query,k=k,filter=filter)
    
    def _split_batch_and_add(self,docs,batch,func) ->  list[list]:
        # ["a","b","c","d","e"] -> [["a","b"],["c","d"],["e"]] when batch is 3
        batch_items = []
        n = len(docs)
        # 特殊情况处理
        if batch <= 0:
            pass  # 如果batch为0或负数，返回空数组
        elif batch == 1:
            batch_items=docs  # 如果batch为1，返回原数组的单元素数组
        elif batch >= n:
            batch_items =  [[docs[i]] for i in range(n)]  # 如果batch大于数组长度，返回每个元素的单独数组
        else:
            # 计算每个batch应该包含的基本数量和多余的元素
            base_size = n // batch
            extra = n % batch

            start = 0

            for i in range(batch):
                # 每个batch的大小，如果有多余的元素，则当前batch大小加1
                current_size = base_size + (1 if i < extra else 0)
                end = start + current_size
                batch_items.append(docs[start:end])
                start = end
        print(f"size = {len(docs)},parts = {[len(item) for item in batch_items]}")
        all_ids = []
        with tqdm(total= len(batch_items),desc="文档片段",colour="#9999ff") as pbar:
            for index,items in enumerate(batch_items):
                ids = func(items)
                all_ids.extend(ids)
                pbar.update(1)
        return all_ids



from __future__ import annotations
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ylz_utils.langchain.graph.self_rag_graph import SelfRagGraph

from langchain_core.prompts import ChatPromptTemplate
from langchain_core.documents import Document
from typing import List

class GradeDocumentsNode():
    def __init__(self,self_rag_graph:SelfRagGraph):
        self.self_rag_graph = self_rag_graph
    def get_retrieval_grader(self):
        # Prompt
        system = """You are a grader assessing relevance of a retrieved document to a user question. \n 
            It does not need to be a stringent test. The goal is to filter out erroneous retrievals. \n
            If the document contains keyword(s) or semantic meaning related to the user question, grade it as relevant. \n
            Give a binary score 'yes' or 'no' score to indicate whether the document is relevant to the question."""
        grade_prompt = ChatPromptTemplate.from_messages(
            [
                ("system", system),
                ("human", "Retrieved document: \n\n {document} \n\n User question: {question}"),
            ]
        )

        retrieval_grader = grade_prompt | self.self_rag_graph.normal_llm_grader
        #question = "agent memory"
        #docs = self.self_rag_graph.retriever.get_relevant_documents(question)
        #doc_txt = docs[1].page_content
        #print(retrieval_grader.invoke({"question": question, "document": doc_txt}))        
        return retrieval_grader
    
    def grade_documents(self,state):
        """
        Determines whether the retrieved documents are relevant to the question.

        Args:
            state (dict): The current graph state

        Returns:
            state (dict): Updates documents key with only filtered relevant documents
        """

        print("---CHECK DOCUMENT RELEVANCE TO QUESTION---")
        question = state["question"]
        documents:List[Document] = state["documents"]

        # Score each doc
        filtered_docs = []
        for d in documents:
            score = self.get_retrieval_grader().invoke(
                {"question": question, "document": d.page_content}
            )
            grade = score.binary_score
            if grade == "yes":
                print("---GRADE: DOCUMENT RELEVANT---")
                filtered_docs.append(d)
            else:
                print("---GRADE: DOCUMENT NOT RELEVANT---")
                continue
        return {"documents": filtered_docs, "question": question}


from __future__ import annotations
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from ylz_utils.langchain.vectorstores import VectorstoreLib

from langchain_elasticsearch import ElasticsearchStore,DenseVectorStrategy
from langchain_elasticsearch.client import create_elasticsearch_client
from langchain_core.documents import Document
from tqdm import tqdm
from typing import List


from ylz_utils.config import Config

class ESLib():
    def __init__(self,vectorstoreLib:VectorstoreLib):
        self.config = Config()
        self.vectorstoreLib = vectorstoreLib
        self.client = None
        self.es_host=self.config.get("VECTORSTORE.ES.HOST")
        self.es_user=self.config.get("VECTORSTORE.ES.USER")
        self.es_password=self.config.get("VECTORSTORE.ES.PASSWORD")
    def init_client(self,host=None,es_user=None,es_password=None):
        es_connection = create_elasticsearch_client(
            url=host or self.es_host,
            username=es_user or self.es_user,
            password=es_password or self.es_password,
            params = {"verify_certs":False,"ssl_show_warn":False},
        )
        self.client = es_connection
        return self.client
    
    def get_store(self,index_name="langchain_index",embedding=None) -> ElasticsearchStore:        
        if not embedding:
            embedding = self.vectorstoreLib.langchainLib.get_embedding()
        if not self.client:
            self.init_client()
        return ElasticsearchStore(
                    es_connection=self.client,
                    index_name=index_name,
                    embedding=embedding
                )
    def delete_store(self,index_name):
        # "test-metadata, test-elser, test-basic"
        return self.client.indices.delete(
            index=index_name,
            ignore_unavailable=True,
            allow_no_indices=True
        )
            
    def add_docs(self,vector_store: ElasticsearchStore,docs,batch:int=1) -> list[str]:
        all_ids = self.vectorstoreLib._split_batch_and_add(docs,batch,vector_store.add_documents)
        return all_ids

    def add_texts(self,vector_store: ElasticsearchStore,texts,batch:int=1) -> list[str]:
        all_ids = self.vectorstoreLib._split_batch_and_add(texts,batch,vector_store.add_texts)
        return all_ids
        
    def delete(self,vectorstore: ElasticsearchStore,ids: List[str] | None = None):
        return vectorstore.delete(ids) 
          
    def search(self,query,vectorstore: ElasticsearchStore,k=10,filter=[]):
        return vectorstore.similarity_search(query,k=k,filter=filter)
    
    def search_with_score(self,query,vectorstore: ElasticsearchStore,k=10,filter=[]):
        return vectorstore.similarity_search_with_score(query,k=k,filter=filter)



from __future__ import annotations
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ylz_utils.langchain.graph.self_rag_graph import SelfRagGraph

class RetrieveNode():
    def __init__(self,self_rag_graph:SelfRagGraph):
        self.self_rag_graph = self_rag_graph

    def retrieve(self,state):
        """
        Retrieve documents

        Args:
            state (dict): The current graph state

        Returns:
            state (dict): New key added to state, documents, that contains retrieved documents
        """
        print("---RETRIEVE---")
        question = state["question"]
        print(state)
        # Retrieval
        documents = self.self_rag_graph.retriever.get_relevant_documents(question)

        return {"documents": documents, "question": question}



from typing import Optional,Union,Literal
from ylz_utils.config import Config
import logging

from neo4j import GraphDatabase, Result
from neo4j.graph import Node,Relationship

from ylz_utils.data import StringLib

class Neo4jLib():
    def __init__(self,host:str=None,user:str=None ,password:str=None):
        self.config = Config()
        self.host = host or self.config.get("NEO4J.HOST")
        self.user = user or self.config.get("NEO4J.USER")
        self.password = password or self.config.get("NEO4J.PASSWORD")
        self.get_driver()

    def get_driver(self):
        self.driver = GraphDatabase.driver(self.host,auth=(self.user,self.password))
        return self.driver
    
    def close(self):
        if self.driver:
            self.driver.close()

    def run(self,command:str,**kwargs) -> Result:
        if not self.driver:
            self.driver = self.get_driver()
        with self.driver.session(database="neo4j") as session:
            try:
                return session.run(command,**kwargs)
            except Exception as e:
                logging.error(e)
                raise e
    
    def query(self,query:str,only_data=False,**kwargs):
        if not self.driver:
            self.get_driver()
        try:
            records,summary,keys = self.driver.execute_query(query,database_="neo4j",**kwargs) 
            if only_data :
                return [ record.data() for record in records]
            else:
                return records,summary,keys   
        except Exception as e:
            logging.error(e)
            raise e
    def create_relationships(self,*,objects=None,subjects=None,relationships,object_label,subject_label=None,key:str="name"):
        if not object_label:
            raise Exception("必须定义object_label!")
        if not subject_label:
            subject_label = object_label
            print(f"when subject_label is None,then subject_label=object_label={{{object_label}}}")
        '''
        创建主体节点(objects)和客体节点(subjects),主体和客体节点可以重复,以key为唯一标识。
          - 如果objects存在则merge方式创建objects
          - 如果subjects存在则merge方式创建subjects
          - objects与subjects的第一行属性作为节点属性,每一行都应保持同样的结构
          - object_label必须指定
          - 如果subject_label未指定则默认与object_label相同
        创建关系根据relationships,为(object)-[type]->(subject)关系表
          - 如果relationships存在则merge方式创建relation
          - 每一行结构必须包含object,subject,type字段，分别代表主体，客体和关系
          - 第一行属性(除object,subject,type外)作为关系属性，相同type的每一行都应保持同样的结构
          
        例如: 以下数据,key为name
        # 定义Person数据
        objects = [
            {"name": "Alice", "age": 30},
            {"name": "Bob", "age": 25},
            {"name": "Charlie", "age": 35}
        ]
        subjects = [
            {"name": "Alice", "age": 30},
            {"name": "Bob", "age": 25},
            {"name": "Charlie", "age": 35}
        ]

        # 定义关系数据
        relationships = [
            {"object": "Alice", "subject": "Bob","type":"朋友"},
            {"object": "Alice", "subject": "Charlie","type":"领导"},
            {"object": "Bob", "subject": "Charlie","type":"同事"}
        ]
        create_relationships(objects,subjects,relationships,object_label='Person',subject_label='Person',key='name')
        '''
        with self.driver.session() as session:
            # 批量创建节点和关系
            if objects:
                object_properties = "{" + ",".join([f"{key}:nodea.{key}" for key in objects[0]]) + "}"
                create_objects = f"""
                UNWIND $objects AS nodea
                MERGE (n:{object_label} {object_properties})
                """
                print(StringLib.lred(create_objects))
                session.run(create_objects, objects=objects)
            if subjects:            
                subject_properties = "{" + ",".join([f"{key}:nodeb.{key}" for key in subjects[0]]) + "}"
                create_subjects = f"""
                UNWIND $subjects AS nodeb
                MERGE (m:{subject_label} {subject_properties})
                """
                print(StringLib.green(create_subjects))
                session.run(create_subjects,subjects=subjects)
            
            if relationships:
                group_relation_types = set([ item["type"] for item in relationships ])
                group_relations =[]
                for relation_type in group_relation_types:
                    group_relations.append([item for item in relationships if item["type"]==relation_type])
                
                for relations in group_relations:
                    relation_type = relations[0]["type"]
                    relation_prop = [ key for key in relations[0].keys() if key not in ['object','subject','type']]
                    relation_properties = "{" + ",".join([f"{key}:rel.{key}" for key in relation_prop]) + "}"
                    
                    create_relationships = f"""
                      UNWIND $relations AS rel
                      MATCH (a:{object_label} {{{key}: rel.object}}), (b:{subject_label} {{{key}: rel.subject}})
                      MERGE (a)-[:{relation_type} {relation_properties}]->(b)
                    """
                    print(StringLib.yellow(create_relationships))
                    session.run(create_relationships,relations=relations)


import re
from .state import AgentState


def extract_python_code(text):
    pattern = r'```python\s*(.*?)\s*(```|$)'
    matches = re.findall(pattern, text, re.DOTALL)
    return matches


error_parsing = """Make sure your response contains a code block in the following format:

```python
...
```

When trying to parse out that code block, got this error: {error}"""


def check(state: AgentState):
    last_answer = state['messages'][-1]
    try:
        code_blocks = extract_python_code(last_answer.content)
    except Exception as e:
        return {"messages": [{"role": "user", "content": error_parsing.format(error=str(e))}]}
    if len(code_blocks) == 0:
        return {"messages": [{"role": "user", "content": error_parsing.format(error="Did not find a code block!")}]}
    if len(code_blocks) > 1:
        return {"messages": [{"role": "user", "content": error_parsing.format(error="Found multiple code blocks!")}]}
    return {"code": f"```python\n{code_blocks[0][0]}\n```"}



import ipfshttpclient2
from ipfshttpclient2 import DEFAULT_ADDR,DEFAULT_BASE

class IpfsLib():    
    def __init__(self,addr:str=DEFAULT_ADDR,base:str=DEFAULT_BASE):
        print("正在初始化IPFS...")
        self.client = ipfshttpclient2.connect(addr,base)
        self.add_bytes = self.client.add_bytes
        self.add_json = self.client.add_json
        self.add_str = self.client.add_str
        self.add = self.client.add
        self.get = self.client.get
        self.get_json = self.client.get_json
        self.id = self.client.id
        self.cat = self.client.cat
        self.ls = self.client.ls
        self.pubsub = self.client.pubsub
        self.key = self.client.key



    


from __future__ import annotations
from typing import TYPE_CHECKING, Union
if TYPE_CHECKING:
    from ylz_utils.langchain import LangchainLib

from http import HTTPStatus
from urllib.parse import urlparse, unquote
from pathlib import PurePosixPath
import requests
from dashscope import ImageSynthesis

class FluxLib():
    def __init__(self,langchainLib:LangchainLib):
        self.langchainLib = langchainLib
        self.config = langchainLib.config
    def init(self):
        api_keys = self.config.get('FLUX.DASHSCOPE.API_KEYS')        
        self.model = self.config.get('FLUX.DASHSCOPE.MODEL') or 'flux-schnell'
        if api_keys:
            api_key = self.langchainLib.split_keys(api_keys)[0]
            self.api_key=api_key
        else:
            raise Exception("请先设置FLUX.DASHSCOPE.API_KEYS")
    def gen(self,input_prompt) -> str:
        self.init()
        rsp = ImageSynthesis.call(model=self.model,
                                api_key= self.api_key,
                                prompt=input_prompt,
                                size='1024*1024')
        if rsp.status_code == HTTPStatus.OK:
            print(rsp.output)
            print(rsp.usage)
            # save file to current directory
            for result in rsp.output.results:
                file_name = PurePosixPath(unquote(urlparse(result.url).path)).parts[-1]
                with open('./%s' % file_name, 'wb+') as f:
                    f.write(requests.get(result.url).content)
                return file_name
        else:
            print('Failed, status_code: %s, code: %s, message: %s' %
                (rsp.status_code, rsp.code, rsp.message))
            return ""

    def agen(self,input_prompt):
        self.init()
        rsp = ImageSynthesis.async_call(model=self.model,
                                        api_key=self.api_key,
                                        prompt=input_prompt,
                                        size='1024*1024')
        if rsp.status_code == HTTPStatus.OK:
            print(rsp.output)
            print(rsp.usage)
        else:
            print('Failed, status_code: %s, code: %s, message: %s' %
                (rsp.status_code, rsp.code, rsp.message))
        status = ImageSynthesis.fetch(rsp)
        if status.status_code == HTTPStatus.OK:
            print(status.output.task_status)
        else:
            print('Failed, status_code: %s, code: %s, message: %s' %
                (status.status_code, status.code, status.message))

        rsp = ImageSynthesis.wait(rsp)
        if rsp.status_code == HTTPStatus.OK:
            print(rsp.output)
        else:
            print('Failed, status_code: %s, code: %s, message: %s' %
                (rsp.status_code, rsp.code, rsp.message))


if __name__ == '__main__':
    prompt = "Eagle flying freely in the blue sky and white clouds"
    prompt_cn = "一只飞翔在蓝天白云的鹰"
    fluxLib = FluxLib(None)
    fluxLib.gen(prompt)
    fluxLib.agen(prompt_cn)


import functools
from functools import lru_cache
import time
import requests


CACHE_DURATION = 24 * 60 * 60


def time_based_cache(seconds):
    def wrapper_cache(func):
        func = lru_cache(maxsize=None)(func)
        func.lifetime = seconds
        func.expiration = time.time() + func.lifetime

        @functools.wraps(func)
        def wrapped_func(*args, **kwargs):
            if time.time() >= func.expiration:
                func.cache_clear()
                func.expiration = time.time() + func.lifetime
            return func(*args, **kwargs)

        return wrapped_func

    return wrapper_cache


@time_based_cache(CACHE_DURATION)
def load_github_file(url):
    # Convert GitHub URL to raw content URL
    raw_url = url.replace("github.com", "raw.githubusercontent.com").replace("/blob/", "/")

    # Send a GET request to the raw URL
    response = requests.get(raw_url)

    # Check if the request was successful
    if response.status_code == 200:
        return response.text
    else:
        return f"Failed to load file. Status code: {response.status_code}"



from __future__ import annotations
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ylz_utils.langchain.graph import GraphLib

from langgraph.graph import StateGraph, START,END, MessagesState
from langchain_core.messages import AIMessage
from langgraph.graph.state import CompiledStateGraph

from .state import GraphState,GradeDocuments,GradeHallucinations,GradeAnswer

from .node_grade_documents import GradeDocumentsNode
from .node_retrieve import RetrieveNode
from .node_generate import GenerateNode
from .node_transform_query import TransformQueryNode
from .edge_decide_to_generate import DecideToGenerateEdge
from .edge_grade_generation_v_documents_and_question import GradeGenerationVDocumentsAndQuestionEdge


from langchain_community.document_loaders import WebBaseLoader
from langchain_community.vectorstores import Chroma
from langchain_openai import OpenAIEmbeddings

class SelfRagGraph():
    retriever = None
    llm = None
    normal_llm_grader = None
    hallucinations_llm_grader = None
    def __init__(self,graphLib:GraphLib):
        self.graphLib = graphLib
        self.retrieve = RetrieveNode(self).retrieve
        self.grade_documents = GradeDocumentsNode(self).grade_documents
        self.generate = GenerateNode(self).generate
        self.transform_query = TransformQueryNode(self).transform_query
        self.decide_to_generate = DecideToGenerateEdge(self).decide_to_generate
        self.grade_generation_v_documents_and_question = GradeGenerationVDocumentsAndQuestionEdge(self).grade_generation_v_documents_and_question
    
    def set_retriever(self,retriever=None):
        if retriever:
            self.retriever = retriever
        else:
            dbname = "langgraph1.faiss"
            try:
                vectorstore = self.graphLib.langchainLib.vectorstoreLib.faissLib.load(dbname)
            except:
                url = "https://langchain-ai.github.io/langgraph/how-tos/"
                docs = self.graphLib.langchainLib.documentLib.url.load_and_split(url=url,max_depth=1,chunk_size=256,chunk_overlap=0)
                print(f"there is {len(docs)} docs to be add to {dbname}.")
                # Add to vectorDB
                print(docs)
                vectorstore, _= self.graphLib.langchainLib.vectorstoreLib.faissLib.create_from_docs(docs)
                vectorstore.save_local(dbname)
            self.retriever = vectorstore.as_retriever()

    def set_llm(self,llm_key,llm_model):
        self.llm = self.graphLib.langchainLib.get_llm(llm_key,llm_model)
        self.normal_llm_grader = self.llm.with_structured_output(GradeDocuments)
        self.hallucinations_llm_grader = self.llm.with_structured_output(GradeHallucinations)
        self.answer_llm_grader = self.llm.with_structured_output(GradeAnswer)

    def _check(self):
        if not self.retriever:
            raise Exception("先调用set_retriever(retriever)")
        if not self.llm or not self.normal_llm_grader or \
           not self.hallucinations_llm_grader or \
           not self.answer_llm_grader:
            raise Exception("先调用self.set_llm(llm_key,llm_model)")

    def get_graph(self,llm_key,llm_model):
        self.set_llm(llm_key,llm_model)
        self._check()
        workflow = StateGraph(GraphState)

        # Define the nodes
        workflow.add_node("retrieve", self.retrieve)  # retrieve
        workflow.add_node("grade_documents", self.grade_documents)  # grade documents
        workflow.add_node("generate", self.generate)  # generatae
        workflow.add_node("transform_query", self.transform_query)  # transform_query

        # Build graph
        workflow.add_edge(START, "retrieve")
        workflow.add_edge("retrieve", "grade_documents")
        workflow.add_conditional_edges(
            "grade_documents",
            self.decide_to_generate,
            {
                "transform_query": "transform_query",
                "generate": "generate",
            },
        )
        workflow.add_edge("transform_query", "retrieve")
        workflow.add_conditional_edges(
            "generate",
            self.grade_generation_v_documents_and_question,
            {
                "not supported": "generate",
                "useful": END,
                "not useful": "transform_query",
            },
        )

        graph = workflow.compile(self.graphLib.memory)
        
        return graph


import json
import re
from ylz_utils.database.neo4j import Neo4jLib
from ylz_utils.data import StringLib,Spinner

def query_neo4j(neo4jLib:Neo4jLib,vars,query,only_data=False):
    spinner = Spinner()
    try:
        query_variables =re.findall(r"\$(.+?)\s",query)
        kwargs = {}
        for key in query_variables:
            if key in vars:
                kwargs[key] = vars[key]
        spinner.start()
        results = neo4jLib.query(query,only_data,**kwargs)
        spinner.end()
        return results
    except Exception as e:
        spinner.end()      
        raise e
    
def neo4j_test(args):
    user = args.user
    password = args.password
    host = args.host
    neo4jLib = Neo4jLib(host,user,password)
    print("*"*50,"let's start","*"*50)
    idx = 0
    vars = {}
    while True:
        idx +=1
        query = input(f'{StringLib.color(f"输入语句{idx}: ")}')
        if query.strip().lower().startswith('/'):
            command =  query.strip().lower().split(" ")[0]
            if command not in ["/q","/set","/get","/list","/clear","/load","/query"]:
                print(
"""
usage:
    /set <key>=<value> 设置变量
    /get <key> 查看key变量
    /list 变量列表
    /clear 清除变量
    /load <file name> as <key> 读取文件内容设置为变量key
    /query <key>=<CQL查询语句> 将查询语句的结果设置为变量key
    /q 退出
"""
                )
                continue
        if query.strip().lower() == '/q':
            print("Good bye!")
            break
        if query.strip().startswith('/set '):
            kv = query.strip().replace("/set","").strip().split('=')
            try:
                key =kv[0].strip()
                value = kv[1].strip()
                vars[key] = json.loads(value)
            except Exception as e:
                print(e)
            continue
        if query.strip().startswith('/get '):
            key = query.strip().replace("/get","").strip()
            if key in vars:
                print(f"{key}={vars[key]}")
            continue
        if query.strip().lower().startswith('/list'):
            print(list(vars.keys()))
            continue
        if query.strip().lower().startswith('/clear'):
            vars = {}
            continue
        if query.strip().startswith('/query '):
            try:
                commands = query.strip().replace('/query ',"").split("=",1)
                key = commands[0].strip()
                query = commands[1].strip()            
                records = query_neo4j(neo4jLib,vars,query,True)
                vars[key] = records
            except Exception as e:
                print(StringLib.lred("ERROR ON QUERY:"),e)
            continue
        if query.strip().startswith('/load'):
            try:
                commands = query.strip().replace('/load',"").split(" as ")
                filename = commands[0].strip()
                key = commands[1].strip()
                with open(filename,"r") as f:
                    text = f.read()
                    vars[key] = json.loads(text)
            except Exception as e:
                print(StringLib.lred("ERROR ON LOAD:"),e)
            continue
        try:
            records,summary,keys = query_neo4j(neo4jLib,vars,query,False)
            # Records
            for record in records[:5]:
                print(record.data())
            if records[5:]:
                n = len(records) - 5
                if n > 0:
                    print(".........")
                    for record in records[-min(n,5):]:
                        print(record.data())
            # Summary information
            if summary:
                print("The query `{query}` returned {records_count} records in {time} ms.".format(
                    query=StringLib.color(summary.query), records_count=StringLib.color(len(records)),
                    time=StringLib.color(summary.result_available_after)
                ))
        except Exception as e:
            print(StringLib.lred("ERROR ON QUERY:"),e)
        print("\n")

    neo4jLib.close()


from __future__ import annotations
from typing import TYPE_CHECKING, Tuple
if TYPE_CHECKING:
    from ylz_utils.langchain.vectorstores import VectorstoreLib

import faiss
from langchain_community.vectorstores import FAISS
from langchain_community.docstore.in_memory import InMemoryDocstore
from langchain_core.documents import Document

from typing import List
from uuid import uuid4

class FaissLib():
    def __init__(self,vectorstoreLib:VectorstoreLib):
        self.vectorstoreLib = vectorstoreLib
        self.config = self.vectorstoreLib.langchainLib.config
        self.db_file = self.config.get("VECTORSTORE.FAISS.DB_FILE")
        self.collection_name = "index"
    def get_store(self,collection_name=None,embedding=None) -> FAISS:
        if collection_name:
            self.collection_name = collection_name
        if not embedding:
            embedding = self.vectorstoreLib.langchainLib.get_embedding()
        index = faiss.IndexFlatL2(len(embedding.embed_query("hello world")))
        vector_store = FAISS(
            embedding_function=embedding,
            index=index,
            docstore=InMemoryDocstore(),
            index_to_docstore_id={},
        )
        return vector_store
            
    def add_docs(self,vector_store:FAISS,docs,batch:int=1) -> list[str]:
        all_ids = self.vectorstoreLib._split_batch_and_add(docs,batch,vector_store.add_documents)
        return all_ids
    
    def add_texts(self,vector_store:FAISS,texts,batch:int=1) -> list[str]:
        all_ids = self.vectorstoreLib._split_batch_and_add(texts,batch,vector_store.add_texts)
        return all_ids
    
    def delete(self,vectorstore: FAISS,ids: List[str] | None = None):
        return vectorstore.delete(ids) 
      
    def save(self,  vectorstore: FAISS,*,db_file:str=None,collection_name:str = None):
        if not db_file:
            db_file = self.db_file
        if not collection_name:
            collection_name = self.collection_name
        vectorstore.save_local(db_file,collection_name)

    def load(self, db_file:str=None , *, embedding=None, collection_name:str = None) -> FAISS:
        if not db_file:
            db_file = self.db_file
        if not collection_name:
            collection_name = self.collection_name
        if not embedding:
            embedding = self.vectorstoreLib.langchainLib.get_embedding()
        vectorestore = FAISS.load_local(db_file, embeddings=embedding, index_name=collection_name, allow_dangerous_deserialization=True)
        return vectorestore
    
    def search(self,query,vectorstore: FAISS,k=10,filter={}):
        return vectorstore.similarity_search(query,k=k,filter=filter)
    
    def search_with_score(self,query,vectorstore: FAISS,k=10,filter={}):
        return vectorstore.similarity_search_with_score(query,k=k,filter=filter)



from __future__ import annotations
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ylz_utils.langchain.graph import GraphLib

from typing import Literal

from langgraph.graph import StateGraph, END, MessagesState
from langchain_core.messages import AIMessage
from langgraph.graph.state import CompiledStateGraph

from .check import check
from .critique import critique
from .draft import draft_answer
from .gather_requirements import gather_requirements
from .state import AgentState, OutputState, GraphConfig

class EngineerGraph():
    def __init__(self,graphLib:GraphLib):
        self.graphLib = graphLib
        # self.node_llms = {
        #     "draft_answer": {"llm_key": "LLM.GROQ","llm_model":None},
        #     "gather_requirements": {"llm_key": "LLM.GROQ","llm_model":None},
        #     "critique": {"llm_key": "LLM.GROQ","llm_model":None},
        # }
        self.node_llms = None
        self.llm_key = None
        self.llm_model = None
    def route_critique(self,state: AgentState) -> Literal["draft_answer", END]:
        if state['accepted']:
            return END
        else:
            return "draft_answer"
    def route_check(self, state: AgentState) -> Literal["critique", "draft_answer"]:
        if isinstance(state['messages'][-1], AIMessage):
            return "critique"
        else:
            return "draft_answer"
    def route_start(self, state: AgentState) -> Literal["draft_answer", "gather_requirements"]:
        if state.get('requirements'):
            return "draft_answer"
        else:
            return "gather_requirements"
    def route_gather(self, state: AgentState) -> Literal["draft_answer", END]:
        if state.get('requirements'):
            return "draft_answer"
        else:
            return END
    def set_node_llms(self,node_llms):
        self.node_llms = node_llms

    def get_node_llm(self,node_key):
        try:
            node_llm = self.node_llms[node_key] 
            llm_key = node_llm.get("llm_key")
            llm_model = node_llm.get("llm_model")
            return self.graphLib.langchainLib.get_llm(key=llm_key,model=llm_model)
        except:            
            return self.graphLib.langchainLib.get_llm(key = self.llm_key, model = self.llm_model)

    def get_graph(self,llm_key=None,llm_model=None) -> CompiledStateGraph:
        # Define a new graph
        self.llm_key = llm_key
        self.llm_model = llm_model
        workflow = StateGraph(AgentState, input=MessagesState, output=OutputState, config_schema=GraphConfig)
        workflow.add_node(draft_answer)
        workflow.add_node(gather_requirements)
        workflow.add_node(critique)
        workflow.add_node(check)
        workflow.set_conditional_entry_point(self.route_start)
        workflow.add_conditional_edges("gather_requirements", self.route_gather)
        workflow.add_edge("draft_answer", "check")
        workflow.add_conditional_edges("check", self.route_check)
        workflow.add_conditional_edges("critique", self.route_critique)
        graph = workflow.compile(checkpointer=self.graphLib.memory)
        return graph 
    






from __future__ import annotations
from typing import TYPE_CHECKING, Union
if TYPE_CHECKING:
    from ylz_utils.langchain import LangchainLib

import dashscope
from dashscope.audio.tts import SpeechSynthesizer

import dashscope
import sys
from dashscope.api_entities.dashscope_response import SpeechSynthesisResponse
from dashscope.audio.tts import ResultCallback, SpeechSynthesizer, SpeechSynthesisResult

class TTSLib():
    model = None
    def __init__(self,langchainLib:LangchainLib):
        self.langchainLib = langchainLib
        self.config = langchainLib.config
    def init(self):
        api_keys = self.config.get('TTS.DASHSCOPE.API_KEYS')        
        self.model = self.config.get('TTS.DASHSCOPE.MODEL') or 'sambert-zhichu-v1'
        if api_keys:
            api_key = self.langchainLib.split_keys(api_keys)[0]
            dashscope.api_key=api_key
        else:
            raise Exception("请先设置TTS.DASHSCOPE.API_KEYS")

    def tts_save(self,text:str,filename:str):
        self.init()
        import pyaudio
        result = SpeechSynthesizer.call(model=self.model,
                                        text=text,
                                        sample_rate=48000)
        if result.get_audio_data() is not None:
            with open(filename, 'wb') as f:
                f.write(result.get_audio_data())
    def tts_play(self,text:str):
        self.init()
        callback = TTS_Callback()
        import pyaudio
        SpeechSynthesizer.call(model='sambert-zhichu-v1',
                       text=text,
                       sample_rate=48000,
                       format='pcm',
                       callback=callback)

class TTS_Callback(ResultCallback):
    _player = None
    _stream = None

    def on_open(self):
        print('Speech synthesizer is opened.')
        self._player = pyaudio.PyAudio()
        self._stream = self._player.open(
            format=pyaudio.paInt16,
            channels=1, 
            rate=48000,
            output=True)

    def on_complete(self):
        print('Speech synthesizer is completed.')

    def on_error(self, response: SpeechSynthesisResponse):
        print('Speech synthesizer failed, response is %s' % (str(response)))

    def on_close(self):
        print('Speech synthesizer is closed.')
        self._stream.stop_stream()
        self._stream.close()
        self._player.terminate()

    def on_event(self, result: SpeechSynthesisResult):
        if result.get_audio_frame() is not None:
            print('audio result length:', sys.getsizeof(result.get_audio_frame()))
            self._stream.write(result.get_audio_frame())

        if result.get_timestamp() is not None:
            print('timestamp result:', str(result.get_timestamp()))       


from .loader import load_github_file
from .llm import _get_llm
from .state import AgentState
from langchain_core.messages import AIMessage
from langchain_core.pydantic_v1 import BaseModel

critique_prompt = """You are tasked with critiquing a junior developers first attempt at building a LangGraph application. \
Here is a long unit test file for LangGraph. This should contain a lot (but possibly not all) \
relevant information on how to use LangGraph.

<unit_test_file>
{file}
</unit_test_file>

Based on the conversation below, attempt to critique the developer. If it seems like the written solution is fine, then call the `Accept` tool.

Do NOT critique the internal logic of the nodes too much - just make sure the flow (the nodes and edges) are correct and make sense. \
It's totally fine to use dummy LLMs or dummy retrieval steps."""


class Accept(BaseModel):
    logic: str
    accept: bool


def _swap_messages(messages):
    new_messages = []
    for m in messages:
        if isinstance(m, AIMessage):
            new_messages.append({"role": "user", "content": m.content})
        else:
            new_messages.append({"role": "assistant", "content": m.content})
    return new_messages


def critique(state: AgentState, config):
    github_url = "https://github.com/langchain-ai/langgraph/blob/main/libs/langgraph/tests/test_pregel.py"
    file_contents = load_github_file(github_url)
    messages = [
                   {"role": "user", "content": critique_prompt.format(file=file_contents)},
                   {"role": "assistant", "content": state.get('requirements')},

               ] + _swap_messages(state['messages'])
    model = _get_llm(config,"critique").with_structured_output(Accept)
    response = model.invoke(messages)
    accepted = response.accept
    if accepted:
        return {
            "messages": [
                {"role": "user", "content": response.logic},
                {"role": "assistant", "content": "okay, sending to user"}],
            "accepted": True
        }
    else:
        return {
            "messages": [
                {"role": "user", "content": response.logic},
            ],
            "accepted": False
        }



from langgraph.graph import MessagesState
from typing import TypedDict, Literal
class AgentState(MessagesState):
    requirements: str
    code: str
    accepted: bool


class OutputState(TypedDict):
    code: str


class GraphConfig(TypedDict):
    gather_model: Literal['openai', 'anthropic']
    draft_model: Literal['openai', 'anthropic']
    critique_model: Literal['openai', 'anthropic']



from __future__ import annotations
from typing import TYPE_CHECKING,Literal,TypedDict,Annotated
if TYPE_CHECKING:
    from ylz_utils.langchain.graph import GraphLib

from langgraph.prebuilt import ToolNode, tools_condition
from langgraph.graph.message import add_messages
from langgraph.graph import START,END,StateGraph,MessagesState
from langchain_core.messages import SystemMessage,HumanMessage,AIMessage,BaseMessage,ToolMessage

from langchain_core.pydantic_v1 import BaseModel,Field

class State(MessagesState):
    ask_human: bool

class Output(BaseModel):
    originText:str = Field(description="原始文本")
    targetText:str = Field(description="翻译后的文本")
    len_target:int = Field(description="翻译后的文本长度")

class TestGraph():
    llm_key = None
    llm_model= None
    def __init__(self,graphLib:GraphLib):
        self.graphLib = graphLib
    def get_graph(self,llm_key,llm_model):
        self.llm_key = llm_key
        self.llm_model = llm_model
        workflow = StateGraph(State)
        workflow.add_node("robot",self.robot)
        workflow.add_edge(START,"robot")
        workflow.add_edge("robot",END)
        graph = workflow.compile(self.graphLib.memory)
        return graph        
    def robot(self,state:State):
        #outputParser = self.graphLib.langchainLib.get_outputParser(Output)
        #prompt = self.graphLib.langchainLib.get_prompt("把以下文本翻译成中文",outputParser = outputParser)
        prompt = self.graphLib.langchainLib.get_prompt("把以下文本翻译成中文")      
        llm = self.graphLib.langchainLib.get_llm(self.llm_key,self.llm_model)
        chain =  prompt | llm.with_structured_output(Output)
        message = state["messages"][-1]         
        translate_dict = chain.invoke({"input":message.content})
        print(translate_dict)
        
        return {"messages":[AIMessage(content = str(translate_dict.targetText))],"ask_human":False}
    